#ifndef lint
static char *RCSid() { return RCSid("$Id: amavis-milter.c,v 1.1.2.11 2001/04/05 18:55:47 lhecking Exp $"); }
#endif

/* 

   amavis-perl-calling milter
   backport of amavisd version by reniar
   v-0.4

   gwinkless@users.sourceforge.net

*/

/*
 * Add some copyright notice here ...
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#define BUFFLEN 255

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <unistd.h>
#include <errno.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <stdarg.h>
#include <time.h>

#include "libmilter/mfapi.h"

typedef int bool;

#ifndef FALSE
# define FALSE	0
#endif /* ! FALSE */
#ifndef TRUE
# define TRUE	1
#endif /* ! TRUE */

#define TEMPLATE "/amavis-milter-XXXXXXXX"
#define AMAVIS_PERL_PATH "/usr/sbin/amavis"
#define DEBUG_FILE "/tmp/virdebug"

#ifndef HAVE_STRLCPY
# define strlcpy mystrlcpy
static size_t
mystrlcpy(char *dst, const char *src, size_t dstsize)
{
    (void) strncpy(dst, src, dstsize - 1);
    *(dst + dstsize - 1) = '\0';
    return strlen(src);
}
#endif

typedef struct llstrct {
    char *str;
    struct llstrct *next;
} ll;

struct mlfiPriv {
    struct in_addr client_addr;
    char *mlfi_fname;
    FILE *mlfi_fp;
    char *mlfi_envfrom;
    ll mlfi_envto;
    ll *mlfi_thisenvto;
    int mlfi_numto;
};

static void mydebug(const char *, ...);
static char *mymktempdir(char *);

static int debugflag;
static char *mydebugfile;

static void
mydebug(const char *fmt, ...)
{
    if (debugflag) {
	FILE *f;
	int fd;
	time_t tmpt;
	char *stime;
	va_list ap;

	if (mydebugfile) {
	    f = fopen(mydebugfile, "a");
	} else {
	    remove(DEBUG_FILE);
	    if ((fd = open(DEBUG_FILE, O_CREAT | O_EXCL | O_RDWR, 0600)) < 0 || (f = fdopen(fd, "a")) == NULL) {
		fprintf(stderr, "error while opening '%s'\n", DEBUG_FILE);
		return;
	    }
	}
	tmpt = time(NULL);
	stime = ctime(&tmpt);
	fprintf(f, "%d:%s", getpid(), stime);

	va_start(ap, fmt);
	vfprintf(f, fmt, ap);
	va_end(ap);

	fputc('\n', f);
	fclose(f);
	close(fd);
    }
}

static char *
mymktempdir(char *s)
{
    char dirname[BUFFLEN];
    char *stt;
    int count = 0;

    strlcpy(dirname, s, sizeof(dirname));

#ifdef HAVE_MKDTEMP
    return mkdtemp(s);
#else
    /* magic number alert */
    while (count++ < 20) {
# ifdef HAVE_MKTEMP
	stt = mktemp(s);
# else
	stt = strchr(dirname, '-') + 1;
	if (stt) {
	    /* more magic number alert */
	    snprintf(stt, sizeof(dirname) - 1 - (stt - dirname), "%08d", lrand48() / 215);
	    /* This assumes that s in the calling function is the same size
	     * as the local dirname! Beware of malicious coders ;-) */
	    strlcpy(s, dirname, sizeof(dirname));
	    stt = dirname;
	} else {
	    /* invalid template */
	    return NULL;
	}
# endif
	if (stt) {
	    if (!mkdir(s, S_IRWXU)) {
		return s;
	    } else {
		continue;
	    }
	}
    }
    return NULL;
#endif /* HAVE_MKDTEMP */
}

#define MLFIPRIV	((struct mlfiPriv *) smfi_getpriv(ctx))

void
freeenvto(ll * envto)
{

    ll *new;

    while (envto) {
	new = envto->next;
	if (envto->str) {
	    free(envto->str);
	    envto->str = NULL;
	}
	if (envto)
	    free(envto);
	envto = new;
    }
}



sfsistat
clearpriv(ctx, retme, clearall)
SMFICTX *ctx;
sfsistat retme;
int clearall;
{

    /* release private memory and return retme */
    struct mlfiPriv *priv = MLFIPRIV;

    mydebug(" Clearing priv (clearall=%d)\n", clearall);
    if (priv) {
	if (priv->mlfi_fname) {
	    mydebug(" clearing fname\n");
	    free(priv->mlfi_fname);
	    priv->mlfi_fname = NULL;
	}
	if (priv->mlfi_envfrom) {
	    mydebug(" clearing envfrom\n");
	    free(priv->mlfi_envfrom);
	    priv->mlfi_envfrom = NULL;
	}
	if (priv->mlfi_envto.next) {
	    mydebug(" clearing multi-envto\n");
	    freeenvto(priv->mlfi_envto.next);
	    priv->mlfi_envto.next = NULL;
	}
	if (priv->mlfi_envto.str) {
	    mydebug(" clearing envto\n");
	    free(priv->mlfi_envto.str);
	    priv->mlfi_envto.str = NULL;
	}
	if (clearall) {
	    mydebug(" clearing priv\n");
	    free(priv);
	    priv = NULL;
	}
	smfi_setpriv(ctx, priv);
    }
    return retme;
}

extern sfsistat mlfi_cleanup(SMFICTX *, bool);

int
allocmem(SMFICTX * ctx)
{
/*
	allocate some private memory if not already allocated

	returns 0 if ok, 1 if not
*/
    struct mlfiPriv *priv = MLFIPRIV;

    if (priv == NULL) {
	mydebug(" priv was null\n");
	priv = malloc(sizeof *priv);
	if (priv == NULL) {
	    /* can't accept this message right now */
	    fprintf(stderr, "failed to malloc %d bytes for private store\n", sizeof(*priv));
	    return 1;
	}
	mydebug(" malloced priv - now using memset()\n");
	memset(priv, 0, sizeof *priv);
	mydebug(" malloced priv successfully\n");
	smfi_setpriv(ctx, priv);
    } else {
	mydebug(" allocmem tried but priv was already set\n");
	mydebug(" priv->client_addr.s_addr is %d\n", priv->client_addr.s_addr);
    }
    return 0;
}

sfsistat
mlfi_connect(SMFICTX * ctx, char *hostname, _SOCK_ADDR * gen_hostaddr)
{
    struct mlfiPriv *priv;
    struct sockaddr_in *hostaddr;

    hostaddr = (struct sockaddr_in *) gen_hostaddr;

    if (hostaddr) {
	mydebug(" hostname is %s, addr is %d.%d.%d.%d\n", hostname, (hostaddr->sin_addr.s_addr) & 0xff,
		(hostaddr->sin_addr.s_addr >> 8) & 0xff, (hostaddr->sin_addr.s_addr >> 16) & 0xff,
		(hostaddr->sin_addr.s_addr >> 24) & 0xff);
    }
    mydebug(" checking allocmem\n");
    if (allocmem(ctx))
	return SMFIS_TEMPFAIL;
    priv = MLFIPRIV;
    if (hostaddr) {
	priv->client_addr.s_addr = hostaddr->sin_addr.s_addr;
    } else {
	priv->client_addr.s_addr = 0;
    }
    smfi_setpriv(ctx, priv);	/* not really needed but nice */
    return SMFIS_CONTINUE;
}

sfsistat
mlfi_envto(SMFICTX * ctx, char **envto)
{

    struct mlfiPriv *priv;
    if (allocmem(ctx))
	return SMFIS_TEMPFAIL;

    priv = MLFIPRIV;
    if (!(priv->mlfi_thisenvto)) {
/* first one... */
	priv->mlfi_thisenvto = &(priv->mlfi_envto);
	priv->mlfi_numto = 1;
    } else {
	priv->mlfi_numto++;
	if ((priv->mlfi_thisenvto->next = malloc(sizeof(ll))) == NULL)
	    return (SMFIS_TEMPFAIL);
	priv->mlfi_thisenvto = priv->mlfi_thisenvto->next;
	priv->mlfi_thisenvto->next = NULL;
    }
    if ((priv->mlfi_thisenvto->str = strdup(*envto)) == NULL)
	return (SMFIS_TEMPFAIL);
    fprintf(stdout, "%d: added %s as recip\n", getpid(), *envto);
    return SMFIS_CONTINUE;
}

sfsistat
mlfi_envfrom(SMFICTX * ctx, char **envfrom)
{
    struct mlfiPriv *priv;
    char dirname[BUFFLEN];
    char tmpfile[BUFFLEN];
    struct stat buf, StatBuf;

    if (allocmem(ctx))
	return SMFIS_TEMPFAIL;
    priv = MLFIPRIV;

    /* open a file to store this message */
    strlcpy(dirname, RUNTIME_DIR, sizeof(dirname));
    strncat(dirname, TEMPLATE, sizeof(dirname) - 1 - strlen(dirname));
    if (mymktempdir(dirname) == NULL) {
	fprintf(stderr, "Failed to create temp dir\n");
	return SMFIS_TEMPFAIL;
    }
    snprintf(tmpfile, sizeof(tmpfile), "%s/email.txt", dirname);
    fprintf(stdout, "%d: got %s directory, %s sender\n",
	    getpid(), tmpfile, *envfrom);

    priv->mlfi_fname = strdup(tmpfile);
    if (priv->mlfi_fname == NULL) {
	return SMFIS_TEMPFAIL;
    }
    priv->mlfi_envfrom = strdup(*envfrom);
    if (!priv->mlfi_envfrom) {
	free(priv->mlfi_fname);
	priv->mlfi_fname = NULL;
	return SMFIS_TEMPFAIL;
    }
    if (lstat(dirname, &StatBuf) < 0) {
	fprintf(stderr, "Error while trying lstat(%s)\n", dirname);
	exit(SMFIS_TEMPFAIL);
    }

    /* may be too restrictive for you, but's good to avoid problems */
    if (!S_ISDIR(StatBuf.st_mode) || StatBuf.st_uid != getuid() || StatBuf.st_gid != getgid() || !(StatBuf.st_mode & S_IRWXU)) {
	fprintf(stderr, "Security Warning: %s must be a Directory and owned by User %d and Group %d\n"
		"and just read-/write-able by the User and noone else. Exit.\n", dirname, getuid(), getgid());
	exit(EX_TEMPFAIL);
    }
    /* there is still a race condition here if RUNTIME_DIR is writeable by the attacker :-\ */

    if ((priv->mlfi_fp = fopen(priv->mlfi_fname, "w+")) == NULL) {
	free(priv->mlfi_fname);
	priv->mlfi_fname = NULL;
	free(priv->mlfi_envfrom);
	priv->mlfi_envfrom = NULL;
	return SMFIS_TEMPFAIL;
    }

    stat(dirname, &buf);
    /*if (buf.st_uid!=getuid() || (buf.st_mode & 0777) != 0700) {
       fprintf(stderr, "Security alert -- someone changed %s. Uid is %d, mode is %o\n", dirname, buf.st_uid, 0777 & buf.st_mode);
       fclose (priv->mlfi_fp);
       unlink (priv->mlfi_fname);
       unlink (dirname);
       free(priv->mlfi_fname);
       free(priv->mlfi_envfrom);
       free(priv);
       return SMFIS_TEMPFAIL;
       } */


    /* save the private data */

    smfi_setpriv(ctx, priv);

    /* continue processing */
    return SMFIS_CONTINUE;
}

sfsistat
mlfi_header(ctx, headerf, headerv)
SMFICTX *ctx;
char *headerf;
char *headerv;
{
    /* write the header to the log file */
    fprintf(MLFIPRIV->mlfi_fp, "%s: %s\n", headerf, headerv);

    /* continue processing */
    return SMFIS_CONTINUE;
}

sfsistat
mlfi_eoh(ctx)
SMFICTX *ctx;
{
    /* output the blank line between the header and the body */
    fprintf(MLFIPRIV->mlfi_fp, "\n");

    /* continue processing */
    return SMFIS_CONTINUE;
}

sfsistat
mlfi_body(ctx, bodyp, bodylen)
SMFICTX *ctx;
u_char *bodyp;
size_t bodylen;
{
    /* output body block to log file */

    int y, x;

    /* convert crlf to lf */
    for (x = 0, y = 0; x < bodylen;) {
	bodyp[y++] = bodyp[x++];
	if (bodyp[x] == 13 && bodyp[x + 1] == 10)
	    x++;
    }
    bodylen -= (x - y);


    if (fwrite(bodyp, bodylen, 1, MLFIPRIV->mlfi_fp) <= 0) {
	/* write failed */
	(void) mlfi_cleanup(ctx, FALSE);
	return SMFIS_TEMPFAIL;
    }

    /* continue processing */
    return SMFIS_CONTINUE;
}

sfsistat
mlfi_eom(ctx)
SMFICTX *ctx;
{
    struct mlfiPriv *priv = MLFIPRIV;
    int itmp, pid, status;
    if (priv) {
	if (priv->mlfi_fp)
	    fclose(priv->mlfi_fp);
/* close the file so we can run checks on it!!! */

	pid = fork();
	if (pid == -1) {
	    fprintf(stderr, "couldn't fork!\n");
	    return clearpriv(ctx, SMFIS_TEMPFAIL);
	}
	if (pid == 0) {
	    char **args;
	    int x;
	    char *tmpstr;

	    args = malloc(sizeof(*args) * (priv->mlfi_numto + 4));
	    tmpstr = malloc(3 * sizeof(*tmpstr) * (priv->mlfi_numto + 3) + 3);
	    if (args == NULL || tmpstr == NULL)
		return clearpriv(ctx, SMFIS_TEMPFAIL);

	    for (x = 0; x < priv->mlfi_numto; x++) {
		strncpy(&(tmpstr[x * 3 + 9]), "%s ", 3);
	    }
	    strncpy(tmpstr, "%s %s %s ", 9);
	    tmpstr[x * 3 + 9] = '\n';
	    tmpstr[x * 3 + 10] = 0;
	    args[0] = "";
	    args[1] = priv->mlfi_fname;
	    args[2] = priv->mlfi_envfrom;
	    priv->mlfi_thisenvto = &(priv->mlfi_envto);
	    for (x = 0; x < priv->mlfi_numto; x++) {
		args[x + 3] = strdup(priv->mlfi_thisenvto->str);
		priv->mlfi_thisenvto = priv->mlfi_thisenvto->next;
	    }

	    args[x + 3] = NULL;
	    strrchr(priv->mlfi_fname, '/')[0] = 0;
	    /* amavis-perl wants the directory, not the filename */
	    fprintf(stderr, "%d:", getpid());
	    vfprintf(stderr, tmpstr, args);
	    execv(AMAVIS_PERL_PATH, args);

/* for some reason execl is ignoring the second parameter - so I pass 
   the empty string. Anyone any thoughts as to why this happens??? Geoff */
	    exit(127);
	}
	waitpid(pid, &status, 0);

	if (WIFEXITED(status)) {
	    if (WEXITSTATUS(status) == 127) {
		fprintf(stderr, "%d: exec failed\n", pid);
		return clearpriv(ctx, SMFIS_TEMPFAIL);
	    }
	    fprintf(stderr, "%d: return value was %d\n", pid, WEXITSTATUS(status));
	    if (WEXITSTATUS(status) == 99) {
		fprintf(stderr, "%d: discarding mail\n", pid);
		return clearpriv(ctx, SMFIS_DISCARD);
		/* there was a virus. Discard it so it doesn't go bouncing around!! */
	    }
	    if (WEXITSTATUS(status) == 0)
		smfi_addheader(ctx, "X-Virus-Scanned", "by AMaViS-perl11-milter (http://amavis.org/)");

/* if we got any exit status but 0, we didn't check the file... so don't add
   the header. We could return TEMPFAIL instead, if we wanted to be paranoid */

	    return clearpriv(ctx, SMFIS_ACCEPT);
	} else {
	    fprintf(stderr, "%d: which failed\n", getpid());
	    return clearpriv(ctx, SMFIS_TEMPFAIL);
	}
    }
    return mlfi_cleanup(ctx, TRUE);

}

sfsistat
mlfi_close(ctx)
SMFICTX *ctx;

{
    return clearpriv(ctx, SMFIS_ACCEPT, 1);
}

sfsistat
mlfi_abort(ctx)
SMFICTX *ctx;
{
    return mlfi_cleanup(ctx, FALSE);
}

sfsistat
mlfi_cleanup(ctx, ok)
SMFICTX *ctx;
bool ok;
{
    sfsistat rstat = SMFIS_CONTINUE;
    struct mlfiPriv *priv = MLFIPRIV;
    char *p;
    char host[512];

    if (!priv)
	return rstat;

    /* close the archive file */
    if (priv->mlfi_fp != NULL && fclose(priv->mlfi_fp) == EOF) {
	/* failed; we have to wait until later */
	rstat = SMFIS_TEMPFAIL;
	(void) unlink(priv->mlfi_fname);
    } else if (ok) {
	/* add a header to the message announcing our presence */
	if (gethostname(host, sizeof(host) - 1) < 0)
	    strlcpy(host, "localhost", sizeof host);
	p = strrchr(priv->mlfi_fname, '/');
	if (p == NULL)
	    p = priv->mlfi_fname;
	else
	    p++;
    } else {
	/* message was aborted -- delete the archive file */
	(void) unlink(priv->mlfi_fname);
	*(strrchr(priv->mlfi_fname, '/')) = 0;
	rmdir(priv->mlfi_fname);
    }


    /* return status */
    fprintf(stderr, "cleanup called\n");
    return clearpriv(ctx, rstat, 0);
}


struct smfiDesc smfilter = {
    "amavis-milter",		/* filter name */
    SMFI_VERSION,		/* version code -- do not change */
    SMFIF_ADDHDRS,		/* flags */
    mlfi_connect,		/* connection info filter */
    NULL,			/* SMTP HELO command filter */
    mlfi_envfrom,		/* envelope sender filter */
    mlfi_envto,			/* envelope recipient filter */
    mlfi_header,		/* header filter */
    mlfi_eoh,			/* end of header */
    mlfi_body,			/* body block filter */
    mlfi_eom,			/* end of message */
    mlfi_abort,			/* message aborted */
    mlfi_close			/* connection cleanup */
};


int
main(argc, argv)
int argc;
char *argv[];
{
    int c;
    const char *args = "p:d";

    umask(0077);

    debugflag = 0;
    /* Process command line options */
    while ((c = getopt(argc, argv, args)) != -1) {
	switch (c) {
	case 'p':
	    if (optarg == NULL || *optarg == '\0') {
		(void) fprintf(stderr, "Illegal conn: %s\n", optarg);
		exit(EX_USAGE);
	    }
	    (void) smfi_setconn(optarg);
	    break;
	case 'd':
	    debugflag = 1;
	    fprintf(stderr, "debugflag is now 1\n");
	    break;
	}
    }
    if (smfi_register(smfilter) == MI_FAILURE) {
	fprintf(stderr, "smfi_register failed\n");
	exit(EX_UNAVAILABLE);
    }
    return smfi_main();
}

/* eof */
